<?php
// $Id: OptionsFactory.php,v 1.1.4.6 2010/12/27 19:40:13 pounard Exp $

/**
 * This class intends to provide a default IOptions storage mecanism. Notice
 * that all these options will be stored in the same database table, whatever
 * their registry type or type is.
 * 
 * If you use your own factory for business objects that implements this
 * interface, you can bypass this factory and use your own.
 * 
 * For performance reasons, and because we have a totally different schema than
 * the XoXo object table, we are going to use another custom factory
 * mecanism.
 * 
 * No caching at all is provided, neither static, neither binary. This API is
 * meant to be used by high level APIs that should provide their own caching
 * mecanism.
 * By providing our own caching mecanism here, we would duplicate caching on
 * higher level, which would be a total waste of space within the database or
 * binary cache, even in memory itself.
 */
class OptionsFactory
{
  /**
   * Separator for sub arrays in options. This tree handling will allow us
   * to deal only with numeric and string values inside the database value
   * field, and also keep track of parenting inside arrays.
   * 
   * @var char
   */
  const PATH_SEP = '\\';

  /**
   * @var OptionsFactory
   */
  public static $__instance;

  /**
   * Get singleton.
   * 
   * @return OptionsFactory
   */
  public static function getInstance() {
    if (!isset(self::$__instance)) {
      self::$__instance = new OptionsFactory();
    }
    return self::$__instance;
  }

  /**
   * Singleton implementation.
   */
  private function __construct() { }

  /**
   * Find child options using a recursive, yet non performant algorithm. This
   * is the first attempt to make it work, we'll see for optimization later.
   * 
   * @return array
   *   Array of oids containing the full tree, including parent. It will be non
   *   ordered, the only goal of this is to do bulk operations (such as delete).
   */
  protected function _findChildrenOptions($oid, $name = NULL, &$done = array()) {
    // Break circular dependencies.
    if (array_key_exists($oid, $done)) {
      print_r("Breaking recursivity for $oid\n");
      return;
    }
    $done[$oid] = TRUE;

    // Set correct name for array.
    $ret = array($oid => $name ? $name : '');

    $children = db_select('oox_options', 'o')
      ->fields('o', array('oid', 'parent_name'))
      ->condition('parent', $oid)
      ->execute()
      ->fetchAll();

    // Got childrend of this child, continue recursivity.
    if (!empty($children)) {
      foreach ($children as $child) {
        $ret += $this->_findChildrenOptions($child->oid, $child->parent_name, $done);
      }
    }

    // Returned locally found children.
    return $ret;
  }

  /**
   * Internal non-optimized recursive function that will populate arrays of
   * values. 
   */
  protected function _populateArray(&$options, $path, $value) {
    $current = array_shift($path);
    if (!empty($path)) {
      // We manipulate an array therefore we have no need to declare sub
      // arrays, PHP will deal with multidimentional arrays by itself without
      // any notice or errors.
      $this->_populateArray($options[$current], $path, $value);
    }
    else {
      $options[$current] = $value;
    }
  }

  /**
   * Load an IOptionable instance.
   * 
   * @param int $oid
   *   Options instance unique identifier.
   * @param IOptionnable $object = NULL
   *   (optional) If you are trying to load options for an already loaded
   *   instance, pass it here. Notice that no class or registry type checks
   *   will be done.
   *   If you use this parameter, the object won't be statically cached.
   * 
   * @return IOptionable
   *   Fully loaded object. If $object parameter is set, return the exact same
   *   instance.
   * 
   * @throws OoxException
   *   If section does not exists or registry has not been found.
   */
  public function load($oid, IOptionable $object = NULL) {
    $type = NULL;
    $options = array();

    $select = db_select('oox_options')->fields('oox_options')->condition('oid', $oid);
    $data = $select->execute()->fetchObject();

    // No options set for object.
    if (!$data) {
      throw new OoxException("Metadata for options " . $oid . " does not exists");
    }

    // In case no object has been provided, we need to use the given registry to
    // spawn a new instance.
    if (!$object instanceof IOptionable) {
      $registry = OoxRegistry::getInstance($data->registry_type);
      $object = $registry->getItem($data->type);
    }

    // Build the instance.
    $result = db_select('oox_options_data')
      ->fields('oox_options_data')
      ->condition('oid', $oid)
      ->execute();

    // Deal with non IOptionable children by pragmatically and recursively
    // populating the options array.
    while ($row = $result->fetchObject()) {
      $path = explode(OptionsFactory::PATH_SEP, $row->name);
      $this->_populateArray($options, $path, unserialize($row->value));
    }

    // Load children.
    foreach ($this->_findChildrenOptions($oid) as $_oid => $name) {
      if ($name) {
        $path = explode(OptionsFactory::PATH_SEP, $name);
        $this->_populateArray($options, $path, $this->load($_oid));
      }
    }

    // Final object build.
    $object->setOptions($options);
    $object->_setOptionsIdentifier($oid);
    return $object;
  }

  /**
   * Save a single entry into database. This will be called during save
   * transaction.
   */
  protected function _saveValue($oid, $name, $value) {
    if (is_array($value)) {
      foreach ($value as $trail => $value) {
        $this->_saveValue($oid, $name . OptionsFactory::PATH_SEP . $trail, $value);
      }
      // Save the current option as an empty array, in case it's important for
      // implementing high level APIs to fetch back its empty arrays.
      $value = array();
    }

    // Parenting handling.
    else if ($value instanceof IOptionable) {
      $this->save($value, $oid, $name);
      // Do not save further row here.
      return;
    }

    // In most cases, we will get here.
    db_insert('oox_options_data')
      ->fields(array(
        'oid' => $oid,
        'name' => $name,
        'value' => serialize($value),
      ))
      ->execute();
  }

  /**
   * Insert or update an instance. If this instance does not exists in
   * database, its own identifier will be set.
   * 
   * @param IOptionable $options
   *   Section to save.
   * @param array $parent = NULL
   *   (optional) This is an internal parameter. Options can be saved by
   *   cascading IOptionable into one another. This can help saving objects
   *   that provides the IOptionable interface within anothers without having
   *   to deal with cascading save yourself.
   *   This parameter will keep a list of already saved IOptionable interface
   *   identifiers, and will break circular dependencies save attempt.
   * 
   * @throws OoxException
   *   In case of any error.
   */
  public function save(IOptionable $options, $parent = NULL, $parent_name = NULL) {
    // For performance reasons, we wrap all this in a transaction, which will
    // force the table bulk update and indexing. Beside, even if the caller
    // might encapsulate this into their own transaction, this is better for
    // data consistency.
    $transaction = db_transaction();

    try {
      // For performances reasons, we are going to make a BULK insert instead of
      // relying on the drupal_write_record() function.
      if ($oid = $options->getOptionsIdentifier()) {
        db_delete('oox_options_data')
          ->condition('oid', $oid)
          ->execute();
        // Also clear cache.
        unset($this->_cache[$oid]);
      }
      else {
        $type = $options->getType();
        $registryType = $options->getRegistryType();
        // Insert first row, in order to get back new identifier.
        // We don't store the type as a serialized value.
        $oid = db_insert('oox_options')
          ->fields(array(
            'type' => $type,
            'registry_type' => $registryType,
            // Next two can be NULL, this is OK.
            'parent' => $parent,
            'parent_name' => $parent_name,
          ))
          ->execute();
        $options->_setOptionsIdentifier($oid);
      }

      // Save single elements one by one. This might be anti-performance but
      // since I couldn't find out how to do bulk inserts using D7 db layer,
      // we have no other choice. Plus, this algorithm is optimization-free
      // which means simpler to understand.
      foreach ($options->getOptions() as $name => $value) {
        $this->_saveValue($oid, $name, $value);
      }
    }
    // I hope error would not happen here. This shouldn't except in case of
    // buggy algorithm, or database down.
    catch (PDOException $e) {
      $transaction->rollback();
      watchdog_exception('oox_options', $e);
      throw new OoxException("Error during save (or update)");
    }
  }

  /**
   * Delete options for the given options identifier.
   */
  public function delete($oid) {
    try {
      // Find out all options to delete, since we handle parenting.
      $oids = array_keys($this->_findChildrenOptions($oid));

      // Go for a nice transaction.
      $transaction = db_transaction();

      // Delete all.
      db_delete('oox_options_data')
        ->condition('oid', $oids, 'IN')
        ->execute();
      db_delete('oox_options')
        ->condition('oid', $oids, 'IN')
        ->execute();
    }
    // I hope error would not happen here. This shouldn't except in case of
    // buggy algorithm, or database down.
    catch (PDOException $e) {
      $transaction->rollback();
      watchdog_exception('oox_options', $e);
      throw new OoxException("Error during delete");
    }
  }
}
